
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dlfcn.h>

#include <erl_driver.h>
#include <erl_driver_tk.h>
#include <erl_driver_pipelib.h>

extern edtk_debug_flag;

int
main(int argc, char *argv[])
{
    ErlIOVec	*ev;
    void	*libhandle;
    ErlDrvEntry* (*driver_init)(void *);
    ErlDrvEntry	*driver_entry;
    ErlDrvData	driver_data;
    int		*pipe_driver_p;

    edtk_debug_flag = 0;

    if (argc != 2) {
	fprintf(stderr, "usage: %s path-to-driver-shared-lib (pipe version)\r\n", argv[0]);
	exit(1);
    }
    edtk_debug("Attempting to dlopen");
    if ((libhandle = dlopen(argv[1], RTLD_NOW)) == NULL) {
	edtk_debug("dlopen failed");
	perror(dlerror());
	exit(2);
    }

    if ((driver_init = (ErlDrvEntry* (*)(void *)) dlsym(libhandle, "driver_init")) == NULL) {
	fprintf(stderr, "Cannot find 'driver_init', exiting\r\n");
	exit(3);
    }
    if ((driver_entry = (*driver_init)(NULL)) == NULL) {
	fprintf(stderr, "'driver_init' returned NULL\r\n");
	exit(4);
    }
    if ((pipe_driver_p = (int *) dlsym(libhandle, "pipe_driver_p")) == NULL) {
	fprintf(stderr, "Cannot find 'pipe_driver_p', exiting\r\n");
	exit(3);
    }
    *pipe_driver_p = 1;		/* Async calls cause problems for us! */

    if ((*driver_entry->init)() != 0) {
	fprintf(stderr, "Isn't init method supposed to return 0??\r\n");
	exit(5);
    }
    driver_data = (*driver_entry->start)((ErlDrvPort) 0xffffffff, "Hello, pipe driver");

    edtk_debug("Gonna enter main loop");
    while ((ev = read_packet_make_iovec(0)) != NULL) {
	(*driver_entry->outputv)(driver_data, ev);
    }
    edtk_debug("Exited main loop");

    dlclose(libhandle);
    exit(0);
}
