-module('berkeley_db_test').
-define(DRV, berkeley_db_drv).

-include("berkeley_db_drv.hrl").

-export([regression/0, leak_test/0]).

-export([repeat/4]).                            %XXX
repeat(0, _, _, _) ->
    ok;
repeat(N, M, F, A) ->
    apply(M, F, A),
    repeat(N - 1, M, F, A).

regression() ->
    ok = regression2(start),
    io:format("All regression tests via linked-in style PASSED.\n"),
    ok = regression2(start_pipe),
    io:format("All regression tests via pipe style PASSED.\n"),

    io:format("All regression tests PASSED.\n"),
    ok.

leak_test() ->
    {ok, Port1, DBENV1} = ?DRV:start(),
    {ok, 0} = ?DRV:debug(Port1, 0),
    {ok, 0} = ?DRV:env_open(Port1, DBENV1, "",
			    ?DB_CREATE bor ?DB_PRIVATE bor ?DB_INIT_MPOOL, 8#644),
    ok = leak_test(Port1, DBENV1),
    ?DRV:shutdown(Port1),
    io:format("leak_test: linked-in driver test finished\n"),

    {ok, Port2, DBENV2} = ?DRV:start_pipe(),
    {ok, 0} = ?DRV:debug(Port2, 0),
    {ok, 0} = ?DRV:env_open(Port2, DBENV2, "",
			    ?DB_CREATE bor ?DB_PRIVATE bor ?DB_INIT_MPOOL, 8#644),
    ok = leak_test(Port2, DBENV2),
    ?DRV:shutdown(Port2),
    io:format("leak_test: pipe driver test finished\n"),

    ok.

regression2(StartFunc) ->
    {ok, Port1, DBENV1} = ?DRV:StartFunc(),
    {ok, 1} = ?DRV:debug(Port1, 1),
    %% Do this one right away, else we can get in big trouble
    %% QQQ {ok, 0} = ?DRV:env_remove(Port1, DBENV1, "", 0),
    ?DRV:shutdown(Port1),
    
    io:format("\nPerform various environment tests\n\n"),
    {ok, Port2, DBENV2} = ?DRV:StartFunc(),
    {ok, 1} = ?DRV:debug(Port2, 1),
    ok = env_tests(Port2, DBENV2),
    ?DRV:shutdown(Port2),
    
    io:format("\nPerform DB table operations tests\n\n"),
    {ok, Port3, DBENV3} = ?DRV:StartFunc(),
    {ok, 1} = ?DRV:debug(Port3, 1),
    %% XXX DB 4.0.14 + Linux 2.2.12 pukes if we don't include DB_PRIVATE, in this driver or in a stand-alone C program, arrgh!
    {ok, 0} = ?DRV:env_open(Port3, DBENV3, "",
			    ?DB_CREATE bor ?DB_PRIVATE bor ?DB_INIT_MPOOL, 8#644),
    ok = db_ops_tests(Port3, DBENV3),
    ok = db_ops_tests2(Port3, DBENV3),
    {ok, 0} = ?DRV:env_close(Port3, DBENV3),
    ?DRV:shutdown(Port3),
    
    io:format("\nPerform DB simple cursor tests\n\n"),
    {ok, Port4, DBENV4} = ?DRV:StartFunc(),
    {ok, 1} = ?DRV:debug(Port4, 1),
    %% XXX DB 4.0.14 + Linux 2.2.12 pukes if we don't include DB_PRIVATE, in this driver or in a stand-alone C program, arrgh!
    {ok, 0} = ?DRV:env_open(Port4, DBENV4, "",
			    ?DB_CREATE bor ?DB_PRIVATE bor ?DB_INIT_MPOOL, 8#644),
    ok = cursor_tests(Port4, DBENV4),
    ?DRV:shutdown(Port4),

    io:format("\nPerform DB lock tests\n\n"),
    {ok, Port5, DBENV5} = ?DRV:StartFunc(),
    {ok, 1} = ?DRV:debug(Port5, 1),
    ok = lock_tests1(Port5, DBENV5),
    ?DRV:shutdown(Port5),
    {ok, Port6, DBENV6} = ?DRV:StartFunc(),
    {ok, 1} = ?DRV:debug(Port6, 1),
    ok = lock_tests2(Port6, DBENV6),
    ?DRV:shutdown(Port6),

    ok.

env_tests(P, DBENV) ->
    %% We only support one environment per port
    {error, _} = ?DRV:env_create(P, 0),

    %% DB_ENV->set_data_dir must be called before DB_ENV->open
    {ok, 0} = ?DRV:env_set_data_dir(P, DBENV, "."),
    {ok, 0} = ?DRV:env_set_shm_key(P, DBENV, 42),
    %% XXX DB 4.0.14 + Linux 2.2.12 pukes if we don't include DB_PRIVATE, in this driver or in a stand-alone C program, arrgh!
    {ok, 0} = ?DRV:env_open(P, DBENV, "",
			    ?DB_CREATE bor ?DB_PRIVATE bor ?DB_INIT_MPOOL, 8#644),
    ok = env_tests_misc(P, DBENV),
    %% We've set various weird env flags, don't use this environment now!
    {ok, 0} = ?DRV:env_close(P, DBENV),
    {error, badarg} = ?DRV:env_close(P, DBENV),
    
    ok.

env_tests_misc(P, DBENV) ->
    %% XXX Won't test DB_ENV->set_errfile yet.

    ok = ?DRV:env_set_errpfx(P, DBENV, "Bummer, dude"),
    {ok, 0} = ?DRV:env_set_flags(P, DBENV,
				 ?DB_NOLOCKING bor ?DB_NOMMAP, 1),
    %% We didn't set up the initial port open with RPC support, so
    %% we expect this to fail.
    {error, _} = ?DRV:env_set_rpc_server(P, DBENV, "localhost", 5, 5, 0),
    {ok, 0} = ?DRV:env_set_tas_spins(P, DBENV, 42),
    {ok, 0} = ?DRV:env_set_tmp_dir(P, DBENV, "."),
    {ok, 0} = ?DRV:env_set_verbose(P, DBENV,
				   ?DB_VERB_WAITSFOR, 1),
    {ok, _Bin1} = ?DRV:db_strerror(P, 1),
    {ok, {_Major, _Minor, _Patch, _Bin2}} =
	?DRV:db_version(P),

    ok.

db_ops_tests(P, DBENV) ->    
    io:format("XXX ZZZ ----- Open DB1\n"),
    {ok, DB1} = ?DRV:db_create(P, DBENV, 0),

    %% Can't do these without an opened table.
    {error, _} = ?DRV:db_del(P, DB1, void, "foo-key", 0),
    {error, _} = ?DRV:db_fd(P, DB1),
    {error, _} = ?DRV:db_get_byteswapped(P, DB1),
    {error, _} = ?DRV:db_get_type(P, DB1),

    %% Open a database on-disk
    Name1 = "regression0.db",
    file:delete(Name1),
    {ok, 0} = ?DRV:db_open(P, DB1, Name1, "", ?DB_BTREE, ?DB_CREATE, 8#644),

    %% Open an in-memory database
    {ok, DB2} = ?DRV:db_create(P, DBENV, 0),
    {ok, 0} = ?DRV:db_open(P, DB2, "", "", ?DB_HASH, ?DB_CREATE, 0),
    {ok, _} = ?DRV:db_fd(P, DB2),		% Huh, it works!
    {ok, 0} = ?DRV:db_close(P, DB2),
    {error, _} = ?DRV:db_close(P, DB2),

    %% Try again, now that the table is open.
    {error, ?DB_NOTFOUND} = ?DRV:db_del(P, DB1, void, "foo-key", 0),
    {ok, _} = ?DRV:db_fd(P, DB1),
    {ok, _} = ?DRV:db_get_byteswapped(P, DB1),
    {ok, _} = ?DRV:db_get_type(P, DB1),

    %% Simple put & get ops
    Key1s = "key1",
    Key1b = list_to_binary(Key1s),
    Val1 = <<"Hello, world!">>,
    {ok, 0} = ?DRV:db_put(P, DB1, void, Key1b, "data1", 0),
    {ok, 0} = ?DRV:db_put(P, DB1, void, Key1b, Val1, 0),
    {error, ?DB_KEYEXIST} =
	?DRV:db_put(P, DB1, void, Key1b, "data1-blahblah", ?DB_NOOVERWRITE),
    ?DRV:db_get(P, DB1, void, Key1s, <<>>, 0), % Problem child in the past.
    {ok, {Key1b, Val1}} = ?DRV:db_get(P, DB1, void, Key1s, <<>>, 0),
    {error, ?DB_NOTFOUND} = ?DRV:db_get(P, DB1, void, "key0", <<>>, 0),
    {error, ?DB_NOTFOUND} = ?DRV:db_get(P, DB1, void, "key2", <<>>, 0),
    {ok, _} = ?DRV:db_sync(P, DB1, 0),
    
    {ok, 0} = ?DRV:db_close(P, DB1),

    %% Create a database, verify it's there, 
    %% add a couple of records and then truncate the entire database, then
    %% rename it, verify it's renamed
    {ok, DB4a} = ?DRV:db_create(P, DBENV, 0),
    Name4a = "regression-before.db",
    Name4b = "regression-after.db",
    {ok, 0} = ?DRV:db_open(P, DB4a, Name4a, "", ?DB_BTREE, ?DB_CREATE, 8#644),
    {ok, _} = file:read_file_info(Name4a),
    {ok, 0} = ?DRV:db_put(P, DB4a, void, "foo", "foodata", 0),
    {ok, 0} = ?DRV:db_put(P, DB4a, void, "bar", "bardata", 0),
    io:format("XXX ZZZ ----- right before truncate\n"),
    {ok, 2} = ?DRV:db_truncate(P, DB4a, void, 0),
    {ok, 0} = ?DRV:db_close(P, DB4a),
    {ok, DB4b} = ?DRV:db_create(P, DBENV, 0),
    {ok, 0} = ?DRV:db_rename(P, DB4b, Name4a, "", Name4b, 0),
    {error, _} = file:read_file_info(Name4a),
    {ok, _} = file:read_file_info(Name4b),

    %% Remove the renamed database file, then verify it's gone
    {ok, DB3b} = ?DRV:db_create(P, DBENV, 0),
    {ok, 0} = ?DRV:db_remove(P, DB3b, Name4b, "", 0),
    {error, _} = file:read_file_info(Name4b),

    %% QQQ No simple test for db_upgrade, alas.
    
    %% QQQ I'm seeing weird things happen with db_verify.  I don't know
    %% much about how DB->verify is supposed to work, but it doesn't
    %% look healthy, and I don't know if it's my fault, so I'm going to
    %% ignore the matter for now.  When I run it, I see several:
    %% "beam in free(): warning: modified (chunk-) pointer" errors.
    %% Hrm, I'll comment this out for now....
    %%
    %{ok, DB5} = ?DRV:db_create(P, DBENV, 0),
    %{ok, _} = ?DRV:db_verify(P, DB5, Name1, "", "", 0),
    %TmpFile5a = "Verify-output",
    %file:delete(TmpFile5a),
    %{ok, 0} = ?DRV:db_verify(P, DB5, Name1, "", TmpFile5a, 0),

    ok.

db_ops_tests2(P, DBENV) ->    
    {ok, DB6} = ?DRV:db_create(P, DBENV, 0),

    TmpFile0 = "TmpErrFile",
    file:delete(TmpFile0),
    {ok, 0} = ?DRV:db_set_errfile(P, DB6, TmpFile0, "w"),
    {ok, _} = file:read_file_info(TmpFile0),
    {ok, 0} = ?DRV:db_close(P, DB6),
    file:delete(TmpFile0),

    {ok, DB7} = ?DRV:db_create(P, DBENV, 0),
    %% DB->set_cachesize will fail in a shared environment
    {error, _} = ?DRV:db_set_cachesize(P, DB7, 0, 256*1024, 0),
    ok = ?DRV:db_set_errpfx(P, DB7, "Error, dude"),
    {ok, 0} = ?DRV:db_set_flags(P, DB7, ?DB_DUP bor ?DB_DUPSORT bor
				        ?DB_REVSPLITOFF),
    {ok, 0} = ?DRV:db_set_pagesize(P, DB7, 4096),
    {ok, 0} = ?DRV:db_set_bt_minkey(P, DB7, 64),
    {ok, 0} = ?DRV:db_close(P, DB7),

    {ok, DB8} = ?DRV:db_create(P, DBENV, 0),
    {ok, 0} = ?DRV:db_set_re_delim(P, DB8, $:),
    {ok, 0} = ?DRV:db_set_re_len(P, DB8, 256),
    {ok, 0} = ?DRV:db_set_re_pad(P, DB8, 16#20),
    {ok, 0} = ?DRV:db_set_re_source(P, DB8, "/etc/passwd"),
    {ok, 0} = ?DRV:db_close(P, DB7),

    {ok, DB9} = ?DRV:db_create(P, DBENV, 0),
    {ok, 0} = ?DRV:db_set_h_ffactor(P, DB9, 0),
    {ok, 0} = ?DRV:db_set_h_nelem(P, DB9, 1024*1024),
    {ok, 0} = ?DRV:db_set_lorder(P, DB9, 4321),
    {ok, 0} = ?DRV:db_close(P, DB9),

    {ok, DB10} = ?DRV:db_create(P, DBENV, 0),
    {ok, 0} = ?DRV:db_set_q_extentsize(P, DB10, 100),
    {ok, 0} = ?DRV:db_close(P, DB10),

    ok.

cursor_tests(P, DBENV) ->    
    %% Open an in-memory database
    {ok, DB2} = ?DRV:db_create(P, DBENV, 0),
    %% Create an in-memory table
    {ok, 0} = ?DRV:db_set_flags(P, DB2, ?DB_DUP bor ?DB_DUPSORT),
    {ok, 0} = ?DRV:db_open(P, DB2, "", "", ?DB_BTREE, ?DB_CREATE, 0),

    %% Put records into the table so that it isn't completely empty.
    Key1s = "key1",
    Key1b = list_to_binary(Key1s),
    Val1 = <<"Hello, world!">>,
    Key2s = "key2",
    Key2b = list_to_binary(Key2s),
    Val2 = <<"Greetings, planet!">>,
    Key3s = "key3",
    Key3b = list_to_binary(Key3s),
    Val3 = <<"I'm not long for this world, er, table">>,
    %% Add records out of sort order.
    {ok, 0} = ?DRV:db_put(P, DB2, void, Key2b, Val1, 0),
    {ok, 0} = ?DRV:db_put(P, DB2, void, Key2s, Val2, 0), % bin, list no matter
    {ok, 0} = ?DRV:db_put(P, DB2, void, Key1b, Val1, 0),
    {ok, 0} = ?DRV:db_put(P, DB2, void, Key3s, Val3, 0),

    {ok, C1} = ?DRV:db_cursor(P, DB2, void, 0),

    {ok, {Key1b, Val1}} = ?DRV:c_get(P, C1, <<>>, <<>>, ?DB_FIRST),
    {ok, 1} = ?DRV:c_count(P, C1, 0),
    {ok, {Key1b, Val1}} = ?DRV:c_get(P, C1, Key1b, <<>>, ?DB_SET),
    {ok, 1} = ?DRV:c_count(P, C1, 0),
    {ok, {Key2b, Val2}} = ?DRV:c_get(P, C1, Key2b, <<>>, ?DB_SET),
    {ok, 2} = ?DRV:c_count(P, C1, 0),
    {ok, {Key2b, Val1}} = ?DRV:c_get(P, C1, <<>>, <<>>, ?DB_NEXT_DUP),
    {error, _} = ?DRV:c_get(P, C1, <<>>, <<>>, ?DB_NEXT_DUP),
    
    {ok, {Key3b, Val3}} = ?DRV:c_get(P, C1, <<>>, <<>>, ?DB_LAST),
    {ok, {Key2b, Val1}} = ?DRV:c_get(P, C1, <<>>, <<>>, ?DB_PREV),
    {ok, {Key2b, Val2}} = ?DRV:c_get(P, C1, <<>>, <<>>, ?DB_PREV),
    {ok, {Key1b, Val1}} = ?DRV:c_get(P, C1, <<>>, <<>>, ?DB_PREV),
    {error, _} = ?DRV:c_get(P, C1, <<>>, <<>>, ?DB_PREV),
    {ok, {Key1b, Val1}} = ?DRV:c_get(P, C1, <<>>, <<>>, ?DB_CURRENT),
    {ok, {Key2b, Val2}} = ?DRV:c_get(P, C1, <<>>, <<>>, ?DB_NEXT),

    {ok, C2} = ?DRV:c_dup(P, C1, 0),
    {ok, {Key2b, Val1}} = ?DRV:c_get(P, C1, <<>>, <<>>, ?DB_NEXT),
    % C2 is currently uninitialized
    {error, _} = ?DRV:c_get(P, C2, <<>>, <<>>, ?DB_CURRENT),
    % DB_NEXT on an uninitialized cursor will return the first key/value
    {ok, {Key1b, Val1}} = ?DRV:c_get(P, C2, <<>>, <<>>, ?DB_NEXT),
    {ok, {Key2b, Val2}} = ?DRV:c_get(P, C2, <<>>, <<>>, ?DB_NEXT),
    {ok, 0} = ?DRV:c_close(P, C2),

    {ok, {Key2b, Val2}} = ?DRV:c_get(P, C1, Key2b, <<>>, ?DB_SET),
    {ok, 0} = ?DRV:c_put(P, C1, Key2b, Val2, ?DB_CURRENT), % Same stuff
    NewVal2 = "Blarf",
    NewVal2b = list_to_binary(NewVal2),
    {ok, 0} = ?DRV:c_put(P, C1, Key2b, NewVal2, ?DB_KEYFIRST),
    {error, _} = ?DRV:c_put(P, C1, Key2b, NewVal2b, ?DB_CURRENT bor ?DB_NODUPDATA),
    {error, _} = ?DRV:c_put(P, C1, Key2b, "Can't clobber current when using sorting func", ?DB_CURRENT),

    {ok, {Key3b, Val3}} = ?DRV:c_get(P, C1, Key3b, <<>>, ?DB_SET),
    {ok, 0} = ?DRV:c_del(P, C1, 0),
    {error, _} = ?DRV:c_get(P, C1, Key3b, <<>>, ?DB_SET),

    {ok, 0} = ?DRV:c_close(P, C1),

    {ok, 0} = ?DRV:db_close(P, DB2),
    ok.

lock_tests2(P, DBENV) ->    
    {ok, 0} = ?DRV:set_lk_conflicts(P, DBENV, <<0, 0, 0, 0>>, 2),
    {ok, 0} = ?DRV:set_lk_detect(P, DBENV, ?DB_LOCK_DEFAULT),
    {ok, 0} = ?DRV:set_lk_max_lockers(P, DBENV, 500),
    {ok, 0} = ?DRV:set_lk_max_locks(P, DBENV, 500),
    {ok, 0} = ?DRV:set_lk_max_objects(P, DBENV, 500),
    ok.
    
lock_tests1(P, DBENV) ->    
    %% XXX DB 4.0.14 + Linux 2.2.12 pukes if we don't include DB_PRIVATE, in this driver or in a stand-alone C program, arrgh!
    {ok, 0} = ?DRV:env_open(P, DBENV, "",
          		    ?DB_CREATE bor ?DB_PRIVATE bor ?DB_INIT_LOCK, 8#644),
    {ok, 0} = ?DRV:lock_detect(P, DBENV, 0, ?DB_LOCK_DEFAULT),
    {ok, LockID0} = ?DRV:lock_id(P, DBENV),
    {ok, LockID1} = ?DRV:lock_id(P, DBENV),
    {ok, LockID2} = ?DRV:lock_id(P, DBENV),
    {ok, LockID3} = ?DRV:lock_id(P, DBENV),
    {ok, LockID4} = ?DRV:lock_id(P, DBENV),
    {ok, 0} = ?DRV:lock_id_free(P, DBENV, LockID0),
    {error, _} = ?DRV:lock_id_free(P, DBENV, LockID0),
    
    {ok, Lock0} = ?DRV:lock_get(P, DBENV, LockID1, 0, 
				"Foo lock0!", ?DB_LOCK_READ),
    {ok, Lock1} = ?DRV:lock_get(P, DBENV, LockID2, 0, 
				"Foo lock0!", ?DB_LOCK_READ),
    io:format("\n\nXXX Lock0 = ~w\n\n", [Lock0]),
    {ok, 0} = ?DRV:lock_put(P, DBENV, Lock0),
    {error, _} = ?DRV:lock_put(P, DBENV, Lock0),

    %% QQQ I don't understand why this fails:
    %% "my_debug_errcall: (null): DB_ENV->lock_get: invalid lock mode 2"
    %% It does the same thing for any lockmode "greater" than DB_LOCK_READ:
    %% write, iwrite, iread, and iwr.
    %{ok, Lock2} = ?DRV:lock_get(P, DBENV, LockID3, 0, 
    %   			"Foo writelock", ?DB_LOCK_WRITE),

    {ok, 0} = ?DRV:env_close(P, DBENV),
    ok.

leak_test(P, DBENV) ->
    %% Open an in-memory database
    {ok, DB2} = ?DRV:db_create(P, DBENV, 0),
    {ok, 0} = ?DRV:db_open(P, DB2, "", "", ?DB_HASH, ?DB_CREATE, 0),
    TheKey = list_to_binary("key"),
    TheValue = list_to_binary(lists:seq(0, 255)),
    Put = fun () ->
		  {ok, 0} = ?DRV:db_put(P, DB2, void, TheKey, TheValue, 0)
	  end,
    ok = do_loop0(99000, Put),
    Get = fun () ->
		  {ok, {TheKey, TheVal}} = ?DRV:db_get(P, DB2, void, TheKey, <<>>, 0)
	  end,
    ok = do_loop0(99000, Get),

    ok.

do_loop0(0, Fun) ->
    ok;
do_loop0(N, Fun) ->
    Fun(),
    do_loop0(N - 1, Fun).
