
-module(gd1_test).
-define(DRV, gd1_drv).

-export([regression/0]).

regression() ->
    {ok, Port} = ?DRV:start(),
    {ok, 1} = ?DRV:debug(Port, 1),
    short_example(Port),
    io:format("All regression tests PASSED.\n"),
    ok.

% Not really a regression test, but rather a direct translation of
% "Here is a short example program" found in "readme.txt" in GD's 
% source distribution.
short_example(Port) ->
    %% Doesn't work, no error checking done by gdImageCreate()!
    %% {error, 0} = ?DRV:gdImageCreate(Port, -1, -2),
    %% Allocate the image: 64 pixels across by 64 pixels tall
    {ok, IM} = ?DRV:gdImageCreate(Port, 64, 64),

    %% Allocate the color black (red, green and blue all minimum)
    %% Since this is the first color in a new image, it will
    %% be the background color.
    {ok, Black} = ?DRV:gdImageColorAllocate(Port, IM, 0, 0, 0),
    %% Allocate the color white (red, green and blue all maximum).
    {ok, White} = ?DRV:gdImageColorAllocate(Port, IM, 255, 255, 255),

    %% Draw a line from the upper left to the lower right,
    %% using white color index.
    ok = ?DRV:gdImageLine(Port, IM, 0, 0, 63, 63, White),
    
    %% Encapsulation-style
    %% Output the image to the disk file in PNG format.
    ok = ?DRV:gdImagePng(Port, IM, "test.png"),
    %% Output the same image in JPEG format, using the default
    %% JPEG quality setting.
    ok = ?DRV:gdImageJpeg(Port, IM, "test.jpg", -1),

    %% Explicit-style ... this time use the fopen- & fclose-style.
    %% Open a file for writing. "wb" means "write binary", important
    %% under MSDOS, harmless under Unix.
    {ok, PngOut} = ?DRV:fopen(Port, "test2.png", "wb"),
    %% Do the same for a JPEG-format file.
    {ok, JpegOut} = ?DRV:fopen(Port, "test2.jpg", "wb"),
    %% Output the image to the disk file in PNG format.
    ok = ?DRV:gdImagePngFILE(Port, IM, PngOut),
    %% Output the same image in JPEG format, using the default
    %% JPEG quality setting.
    ok = ?DRV:gdImageJpegFILE(Port, IM, JpegOut, -1),
    %% Close the files.
    {ok, 0} = ?DRV:fclose(Port, PngOut),
    {ok, 0} = ?DRV:fclose(Port, JpegOut),

    %% Destroy the image in memory.
    ok = ?DRV:gdImageDestroy(Port, IM),
    
    ok.
