
-module(libnet_test).
-define(DRV, libnet_drv).

%% Modify the following to suit your environment:

-define(IFACE_EXIST, "rl0").
-define(IFACE_NOTEXIST, "blar4").
-define(ADDR_MINE, "10.1.1.1").
-define(ADDR_PINGABLE, "10.1.1.254").
%-define(IFACE_EXIST, "eth0").
%-define(IFACE_NOTEXIST, "blar4").
%-define(ADDR_MINE, "10.2.2.10").
%-define(ADDR_PINGABLE, "10.2.2.2").
%-define(IFACE_EXIST, "pcn0").
%-define(IFACE_NOTEXIST, "blar4").
%-define(ADDR_MINE, "10.1.128.33").
%-define(ADDR_PINGABLE, "10.1.128.1").

%% Possible usage: erl +A5 -pz . -s libnet_test regression -s erlang halt

%%
%% In order to verify that the regression test is actually working, you
%% should have a packet capture program (tcpdump, ethereal, etc) sniffing
%% packets on the ?IFACE_EXIST interface.  You should see:
%%
%% 1.  10 ICMP echo request packets sent from ?ADDR_MINE to ?ADDR_PINGABLE
%% 2.  10 ICMP echo reply packets from ?ADDR_PINGABLE in response
%% 3.  1 ICMP echo request packet
%% 3b. Most IP hosts will not respond to the packet from step #3 because
%%     it does not have a correct ICMP checksum.  However, there are some
%%     hosts (for example, my "Graphite" Apple Airport 802.11b wireless
%%     access point/base station) that will respond to it, silly beasts.
%% 4.  1 ICMP echo request packet, this time with a correct ICMP checksum
%% 5.  1 ICMP echo reply packet in response to #4.
%%

-export([regression/0]).

regression() ->
    {ok, Port1} = ?DRV:start(),
    {ok, 1} = ?DRV:debug(Port1, 1),
    r1(Port1),

    {ok, Port2} = ?DRV:start_pipe(),
    {ok, 1} = ?DRV:debug(Port2, 1),
    r1(Port2),

    io:format("All regression tests PASSED.\n"),
    ok.

r1(P) ->
    ok = ?DRV:null(P),

    {ok, S1} = libnet_drv:libnet_open_raw_sock(P, 6),
    
    {ok, 0} = libnet_drv:test_send_ping(P, ?ADDR_MINE, ?ADDR_PINGABLE, S1),

    {ok, 0} = libnet_drv:libnet_close_raw_sock(P, S1),
    {error, badarg} = libnet_drv:libnet_close_raw_sock(P, S1),

    {ok, R1} = libnet_drv:libnet_open_link_interface(P, ?IFACE_EXIST),
    {ok, 1} = libnet_drv:libnet_close_link_interface(P, R1),
    {error, badarg} = libnet_drv:libnet_close_link_interface(P, R1),

    {ok, {_, ?IFACE_EXIST}} = libnet_drv:libnet_select_device(P, ?IFACE_EXIST),
    %% We cannot match on anything specific for my_libnet_select_device_any/1
    %% because if the machine has several interfaces, we don't know which
    %% one libnet will choose.
    {ok, _} = libnet_drv:my_libnet_select_device_any(P),
    {error, _} = libnet_drv:libnet_select_device(P, ?IFACE_NOTEXIST),
    
    {ok, Ping1} =
	libnet_drv:test_make_icmp_echo_no_checksum(P, ?ADDR_MINE,
						   ?ADDR_PINGABLE),
    Ping1Len = size(Ping1),
    {ok, Ping2} = libnet_drv:libnet_do_checksum(P, Ping1, 1, 8),
    Ping2Len = size(Ping2),
    {error, badarg} = libnet_drv:libnet_write_ip(P, S1, Ping1),
    {error, badarg} = libnet_drv:libnet_write_ip(P, S1, Ping2),
    {ok, S2} = libnet_drv:libnet_open_raw_sock(P, 6),
    {ok, Ping1Len} = libnet_drv:libnet_write_ip(P, S1, Ping1),
    {ok, Ping2Len} = libnet_drv:libnet_write_ip(P, S1, Ping2),

    %% Sanity check of contiguousifying lists of binaries!
    <<Ping2a:7/binary, Ping2b:9/binary, Ping2c/binary>> = Ping2,
    {ok, Ping2} = 
	libnet_drv:libnet_do_checksum(P, [Ping2a, Ping2b, Ping2c], 1, 8),


    ok.
