
%%%
%%% To verify that something is happening when this test is run, run
%%% the "spuser" program on localhost or another spread-enabled machine,
%%% then use the command "j foo" to join the "foo" group, then run
%%% this test.
%%%

-module(spread_test).
-define(DRV, spread_drv).

-export([regression/0]).

regression() ->
    {ok, Port1} = ?DRV:start(),
    pathetic_smoke_coverage(Port1),
    ?DRV:shutdown(Port1),

    {ok, Port2} = ?DRV:start_pipe(),
    pathetic_smoke_coverage(Port2),
    ?DRV:shutdown(Port2),

    io:format("All regression tests PASSED.\n"),
    ok.

%%% This little test will generate all of the messages that it
%%% tries to receive.  No external test setup is required.
%%% However, it won't work if the Spread daemon isn't running
%%% on the local machine.
pathetic_smoke_coverage(P) ->
    {ok, PrivName} = ?DRV:sp_connect(P, "", "erl1", 1, 1),
    {ok, 0} = ?DRV:sp_join(P, "foo"),
    {ok, 0} = ?DRV:sp_leave(P, "foo"),
    {ok, 0} = ?DRV:sp_leave(P, "foofoo"),
    {error, -14} = ?DRV:sp_leave(P, "foo#bad"),
    {ok, 13} = ?DRV:sp_multicast(P, 1, "foo", 42, "Hello, world!"),

    %% Avoid race conditions by leaving plenty of time for comm to catch up
    timer:sleep(1000),

    case ?DRV:sp_poll(P) of
	{ok, N} when N > 0 -> ok
    end,
		     
    {ok, _} = ?DRV:sp_receive(P, 4, 400),
    {ok, _} = ?DRV:sp_receive(P, 4, 400),
    %% There should be nothing to receive now
    {ok, 0} = ?DRV:sp_poll(P),

    ok.
